from typing import Optional

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QHBoxLayout, QWidget

from foundry.game.gfx.Palette import PALETTE_GROUPS_PER_OBJECT_SET, PaletteGroup
from foundry.game.level.Level import Level
from foundry.gui.PaletteGroupEditor import PaletteGroupEditor


class PaletteEditor(QWidget):
    palette_updated: SignalInstance = Signal(PaletteGroup, PaletteGroup)  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        bg_palette_group: Optional[PaletteGroup] = None,
        spr_palette_group: Optional[PaletteGroup] = None,
    ):
        super().__init__(parent=parent)
        self._bg_palette_group = (
            bg_palette_group
            if bg_palette_group is not None
            else PaletteGroup(0, 0, [bytearray([0] * 4) for i in range(4)])
        )
        self._spr_palette_group = (
            spr_palette_group
            if spr_palette_group is not None
            else PaletteGroup(0, 0, [bytearray([0] * 4) for i in range(4)])
        )

        layout = QHBoxLayout()
        layout.setSpacing(0)
        self.background_palette_widget = PaletteGroupEditor(self, self.bg_palette_group)
        self.background_palette_widget.palette_updated.connect(lambda *_: self.on_palette_update())
        layout.addWidget(self.background_palette_widget)
        self.sprite_palette_widget = PaletteGroupEditor(self, self.spr_palette_group)
        self.sprite_palette_widget.palette_updated.connect(lambda *_: self.on_palette_update())
        layout.addWidget(self.sprite_palette_widget)
        self.setLayout(layout)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.parent}, {self.bg_palette_group}, {self.spr_palette_group})"

    def load_from_level(self, level: Level):
        self.background_palette_widget.load(level.object_set_number, level.object_palette_index)
        self._bg_palette_group = self.background_palette_widget.palette_group
        self.sprite_palette_widget.load(
            level.object_set_number, level.enemy_palette_index + PALETTE_GROUPS_PER_OBJECT_SET
        )
        self._spr_palette_group = self.sprite_palette_widget.palette_group

    @property
    def changed(self) -> bool:
        return self.bg_palette_group.changed or self.spr_palette_group.changed

    @changed.setter
    def changed(self, value: bool):
        self.bg_palette_group.changed = value
        self.spr_palette_group.changed = value

    @property
    def bg_palette_group(self) -> PaletteGroup:
        return self._bg_palette_group

    @bg_palette_group.setter
    def bg_palette_group(self, palette: PaletteGroup):
        self._bg_palette_group = palette
        self.background_palette_widget.palette_group = palette

    @property
    def spr_palette_group(self) -> PaletteGroup:
        return self._spr_palette_group

    @spr_palette_group.setter
    def spr_palette_group(self, palette: PaletteGroup):
        self._spr_palette_group = palette
        self.sprite_palette_widget.palette_group = palette

    def on_palette_update(self):
        self.palette_updated.emit(self.bg_palette_group, self.spr_palette_group)
        self.changed



from PySide6.QtCore import QRect, QSize
from PySide6.QtGui import QColor, QImage, QPainter, Qt

from foundry.game.gfx.drawable import apply_selection_overlay
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.GraphicsSet import GraphicsSet
from foundry.game.gfx.objects.ObjectLike import ObjectLike
from foundry.game.gfx.Palette import PaletteGroup
from foundry.game.ObjectSet import ObjectSet
from foundry.smb3parse.objects.object_set import (
    ENEMY_ITEM_GRAPHICS_SET,
    ENEMY_ITEM_OBJECT_SET,
)

MASK_COLOR = [0xFF, 0x33, 0xFF]


class EnemyObject(ObjectLike):
    def __init__(self, data, png_data, palette_group: PaletteGroup):
        super(EnemyObject, self).__init__()

        self.is_4byte = False
        self.is_single_block = True
        self.length = 0

        self.obj_index = data[0]
        self.x_position = data[1]
        self.y_position = data[2]

        self.domain = 0

        self.graphics_set = GraphicsSet(ENEMY_ITEM_GRAPHICS_SET)
        self.palette_group = palette_group

        self.object_set = ObjectSet(ENEMY_ITEM_OBJECT_SET)

        self.png_data = png_data

        self.selected = False

        self._setup()

    @property
    def rect(self):
        return QRect(
            self.x_position,
            self.y_position - (self.height - 1),
            self.width,
            self.height,
        )

    def _setup(self):
        obj_def = self.object_set.get_definition_of(self.obj_index)

        self.name = obj_def.description

        self.width = obj_def.bmp_width
        self.height = obj_def.bmp_height

        self._render(obj_def)

    def _render(self, obj_def):
        self.blocks = []

        block_ids = obj_def.object_design

        for block_id in block_ids:
            x = (block_id % 64) * Block.WIDTH
            y = (block_id // 64) * Block.WIDTH

            self.blocks.append(self.png_data.copy(QRect(x, y, Block.WIDTH, Block.HEIGHT)))

    def render(self):
        # nothing to re-render since enemies are just copied over
        pass

    def draw(self, painter: QPainter, block_length, _):
        for i, image in enumerate(self.blocks):
            x = self.x_position + (i % self.width)
            y = self.y_position + (i // self.width)

            y_offset = self.height - 1
            y -= y_offset

            block = image.copy()

            mask = block.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
            block.setAlphaChannel(mask)

            # todo better effect
            if self.selected:
                apply_selection_overlay(block, mask)

            if block_length != Block.SIDE_LENGTH:
                block = block.scaled(block_length, block_length)

            painter.drawImage(x * block_length, y * block_length, block)

    def get_status_info(self):
        return [("Name", self.name), ("X", self.x_position), ("Y", self.y_position)]

    def __contains__(self, item):
        x, y = item

        return self.point_in(x, y)

    def point_in(self, x, y):
        return self.rect.contains(x, y)

    def set_position(self, x, y):
        # todo also check for the upper bounds
        x = max(0, x)
        y = max(0, y)

        self.x_position = x
        self.y_position = y

    def move_by(self, dx, dy):
        new_x = self.x_position + dx
        new_y = self.y_position + dy

        self.set_position(new_x, new_y)

    def get_position(self):
        return self.x_position, self.y_position

    def resize_by(self, dx, dy):
        pass

    @property
    def type(self):
        return self.obj_index

    def change_type(self, new_type):
        self.obj_index = new_type

        self._setup()

    def increment_type(self):
        self.obj_index = min(0xFF, self.obj_index + 1)

        self._setup()

    def decrement_type(self):
        self.obj_index = max(0, self.obj_index - 1)

        self._setup()

    def to_bytes(self):
        return bytearray([self.obj_index, int(self.x_position), int(self.y_position)])

    def as_image(self) -> QImage:
        image = QImage(
            QSize(self.width * Block.SIDE_LENGTH, self.height * Block.SIDE_LENGTH),
            QImage.Format_RGBA8888,
        )

        image.fill(QColor(0, 0, 0, 0))

        painter = QPainter(image)

        self.draw(painter, Block.SIDE_LENGTH, True)

        return image

    def __str__(self):
        return f"{self.name} at {self.x_position}, {self.y_position}"

    def __repr__(self):
        return f"EnemyObject: {self}"



from typing import Callable, Optional

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QVBoxLayout, QWidget

from foundry.game.gfx.Palette import (
    COLORS_PER_PALETTE,
    PALETTES_PER_PALETTES_GROUP,
    PaletteGroup,
    load_palette_group,
)
from foundry.gui.PaletteViewer import PaletteWidget


class PaletteGroupEditor(QWidget):
    palette_updated: SignalInstance = Signal(PaletteGroup)  # type: ignore

    def __init__(self, parent: Optional[QWidget], palette_group: PaletteGroup):
        super().__init__(parent=parent)
        self._palette_group = palette_group

        layout = QVBoxLayout()
        layout.setSpacing(0)
        self.palettes = []
        for short_palette_index in range(PALETTES_PER_PALETTES_GROUP):
            widget = PaletteWidget(self, palette_group, short_palette_index)
            widget.color_changed.connect(self.on_color_update(short_palette_index))
            widget.clickable = True
            self.palettes.append(widget)
            layout.addWidget(widget)

        self.setLayout(layout)

    def load(self, object_set: int, palette_group_index: int):
        self._palette_group = load_palette_group(object_set, palette_group_index)
        for palette in self.palettes:
            palette._palette_group = self._palette_group
            palette._update_colors()

    @property
    def palette_group(self) -> PaletteGroup:
        return self._palette_group

    @palette_group.setter
    def palette_group(self, palette_group: PaletteGroup):
        self._palette_group = palette_group
        for idx in range(PALETTES_PER_PALETTES_GROUP):
            for color in range(COLORS_PER_PALETTE):
                self.palette_group[idx][color] = palette_group[idx][color]

        self.palette_group.changed = True

    def on_color_update(self, short_palette_index: int) -> Callable[[int, int], None]:
        def color_update(index: int, color_index: int):
            if self.palette_group[short_palette_index][index] == color_index:
                return

            # colors at index 0 are shared among all palettes of a palette group
            if index == 0:
                for idx, palette in enumerate(self.palette_group):  # type: ignore
                    palette[0] = color_index
                    self.palettes[idx]._update_colors()
            else:
                self.palette_group[short_palette_index][index] = color_index

            self.palette_group.changed = True
            self.palette_updated.emit(self.palette_group)

        return color_update



from typing import Protocol

from attr import attrib, attrs


class SizeProtocol(Protocol):
    """
    A two dimensional representation of a size of a given object.

    Attributes
    ----------
    width: int
        The width of the object being represented.
    height: int
        The height of the object being represented.
    """

    width: int
    height: int


@attrs(slots=True)
class Size:
    """
    A two dimensional representation of a size, that uses ``attrs`` to create a basic
    implementation.

    Attributes
    ----------
    width: int
        The width of the object being represented.
    height: int
        The height of the object being represented.
    """

    width = attrib(type=int)
    height = attrib(type=int)



from enum import Enum
from json import loads

from pydantic import BaseModel

from foundry import tileset_definitions
from foundry.smb3parse.objects.object_set import (
    AIR_SHIP_OBJECT_SET,
    CLOUDY_OBJECT_SET,
    DESERT_OBJECT_SET,
    DUNGEON_OBJECT_SET,
    ENEMY_ITEM_OBJECT_SET,
    GIANT_OBJECT_SET,
    HILLY_OBJECT_SET,
    ICE_OBJECT_SET,
    MUSHROOM_OBJECT_SET,
    PIPE_OBJECT_SET,
    PIRANHA_PLANT_OBJECT_SET,
    PLAINS_OBJECT_SET,
    SKY_OBJECT_SET,
    SPADE_BONUS_OBJECT_SET,
    UNDERGROUND_OBJECT_SET,
    WATER_OBJECT_SET,
    WORLD_MAP_OBJECT_SET,
)


class GeneratorType(Enum):
    """
    Level objects are generated using different methods, depending on their generator type. Some objects extend until
    they hit another object, some extend up to the sky. To identify in what way a specific type of level object is
    constructed, this enum lists the known generator types.
    """

    HORIZONTAL = 0
    VERTICAL = 1  # vertical downward
    DIAG_DOWN_LEFT = 2
    DESERT_PIPE_BOX = 3
    DIAG_DOWN_RIGHT = 4
    DIAG_UP_RIGHT = 5
    HORIZ_TO_GROUND = 6
    HORIZONTAL_2 = 7  # special case of horizontal, floating boxes, ceilings
    DIAG_WEIRD = 8  #
    SINGLE_BLOCK_OBJECT = 9
    CENTERED = 10  # like spinning platforms
    PYRAMID_TO_GROUND = 11  # to the ground or next object
    PYRAMID_2 = 12  # doesn't exist?
    TO_THE_SKY = 13
    ENDING = 14


class EndType(Enum):
    """
    Some level objects have blocks designated to be used at their ends. For example pipes, which can be extended, but
    always end at one side with the same couple of blocks. To keep track of where those special blocks are to be placed,
    this enum is used. When the value is TWO_ENDS they are always on opposite sides and whether they are left and right
    or top and bottom depends on the generator type of the object.
    """

    UNIFORM = 0
    END_ON_TOP_OR_LEFT = 1
    END_ON_BOTTOM_OR_RIGHT = 2
    TWO_ENDS = 3


ENEMY_OBJECT_DEFINITION = 12


class TilesetDefinition(BaseModel):
    domain: int
    min_value: int
    max_value: int
    bmp_width: int
    bmp_height: int
    object_design: list[int]
    object_design2: list[int]
    rom_object_design: list[int]
    orientation: int
    ending: int
    is_4byte: bool
    description: str

    @property
    def object_design_length(self) -> int:
        return len(self.object_design)


class Tileset(BaseModel):
    __root__: list[TilesetDefinition]


class Tilesets(BaseModel):
    __root__: list[Tileset]


with open(tileset_definitions, "r") as f:
    object_metadata = Tilesets(__root__=loads(f.read()))


object_set_to_definition = {
    WORLD_MAP_OBJECT_SET: 0,
    PLAINS_OBJECT_SET: 1,
    MUSHROOM_OBJECT_SET: 1,
    SPADE_BONUS_OBJECT_SET: 1,
    HILLY_OBJECT_SET: 2,
    SKY_OBJECT_SET: 3,
    DUNGEON_OBJECT_SET: 4,
    AIR_SHIP_OBJECT_SET: 5,
    CLOUDY_OBJECT_SET: 6,
    DESERT_OBJECT_SET: 7,
    WATER_OBJECT_SET: 8,
    PIPE_OBJECT_SET: 8,
    PIRANHA_PLANT_OBJECT_SET: 9,
    GIANT_OBJECT_SET: 9,
    ICE_OBJECT_SET: 10,
    UNDERGROUND_OBJECT_SET: 11,
    ENEMY_ITEM_OBJECT_SET: ENEMY_OBJECT_DEFINITION,
}



from typing import Dict, List, Optional, Tuple

from attr import dataclass
from PySide6.QtGui import QColor

from foundry import root_dir
from foundry.game.File import ROM
from foundry.smb3parse.constants import Palette_By_Tileset, PalSet_Maps
from foundry.smb3parse.levels import BASE_OFFSET

MAP_PALETTE_ADDRESS = PalSet_Maps

PRG_SIZE = 0x2000
PALETTE_PRG_NO = 22

PALETTE_BASE_ADDRESS = BASE_OFFSET + PALETTE_PRG_NO * PRG_SIZE
PALETTE_OFFSET_LIST = Palette_By_Tileset
PALETTE_OFFSET_SIZE = 2  # bytes

PALETTE_GROUPS_PER_OBJECT_SET = 8
ENEMY_PALETTE_GROUPS_PER_OBJECT_SET = 4
PALETTES_PER_PALETTES_GROUP = 4

COLORS_PER_PALETTE = 4
COLOR_SIZE = 1  # byte

PALETTE_DATA_SIZE = (
    (PALETTE_GROUPS_PER_OBJECT_SET + ENEMY_PALETTE_GROUPS_PER_OBJECT_SET)
    * PALETTES_PER_PALETTES_GROUP
    * COLORS_PER_PALETTE
)


@dataclass(eq=False)
class PaletteGroup:
    object_set: int
    index: int
    palettes: List[bytearray]

    changed = False

    def restore(self):
        new_palette_group = load_palette_group(self.object_set, self.index, use_cache=False)

        self.palettes = new_palette_group.palettes

    def __getitem__(self, item):
        return self.palettes[item]

    def __setitem__(self, key, value):
        self.palettes[key] = value

    def __eq__(self, other):
        if not isinstance(other, PaletteGroup):
            raise TypeError(f"Cannot compare PaletteGroup with {type(other)}.")

        return self.object_set == other.object_set and self.index == other.index

    def __hash__(self):
        return hash((self.object_set, self.index))

    def save(self, rom: Optional[ROM] = None):
        if rom is None:
            rom = ROM()

        palette_offset_position = PALETTE_OFFSET_LIST + (self.object_set * PALETTE_OFFSET_SIZE)
        palette_offset = rom.little_endian(palette_offset_position)

        palette_address = PALETTE_BASE_ADDRESS + palette_offset
        palette_address += self.index * PALETTES_PER_PALETTES_GROUP * COLORS_PER_PALETTE

        palettes = []

        for palette in self.palettes:
            palettes.append(rom.write(palette_address, palette))

            palette_address += COLORS_PER_PALETTE


_palette_group_cache: Dict[Tuple[int, int], PaletteGroup] = {}


def load_palette_group(object_set: int, palette_group_index: int, use_cache=True) -> PaletteGroup:
    """
    Basically does, what the Setup_PalData routine does.
    :param object_set: Level_Tileset in the disassembly.
    :param palette_group_index: Palette_By_Tileset. Defined in the level header.
    :param use_cache: Whether to use a cached version, or read from ROM.
    :return: A list of 4 groups of 4 colors.
    """
    key = (object_set, palette_group_index)

    if key not in _palette_group_cache or not use_cache:
        rom = ROM()

        palette_offset_position = PALETTE_OFFSET_LIST + (object_set * PALETTE_OFFSET_SIZE)
        palette_offset = rom.little_endian(palette_offset_position)

        palette_address = PALETTE_BASE_ADDRESS + palette_offset
        palette_address += palette_group_index * PALETTES_PER_PALETTES_GROUP * COLORS_PER_PALETTE

        palettes = []

        for _ in range(PALETTES_PER_PALETTES_GROUP):
            palettes.append(rom.read(palette_address, COLORS_PER_PALETTE))

            palette_address += COLORS_PER_PALETTE

        _palette_group_cache[key] = PaletteGroup(object_set, palette_group_index, palettes)

    return _palette_group_cache[key]


def save_all_palette_groups(rom: Optional[ROM] = None):
    for palette_group in _palette_group_cache.values():
        palette_group.save(rom)

    if rom is None:
        PaletteGroup.changed = False


def restore_all_palettes():
    for palette_group in _palette_group_cache.values():
        palette_group.restore()

    PaletteGroup.changed = False


palette_file = root_dir.joinpath("data", "Default.pal")

with open(palette_file, "rb") as f:
    color_data = f.read()

offset = 0x18  # first color position

NESPalette = []
COLOR_COUNT = 64
BYTES_IN_COLOR = 3 + 1  # bytes + separator

for i in range(COLOR_COUNT):
    NESPalette.append(QColor(color_data[offset], color_data[offset + 1], color_data[offset + 2]))

    offset += BYTES_IN_COLOR


def bg_color_for_object_set(object_set_number: int, palette_group_index: int) -> QColor:
    palette_group = load_palette_group(object_set_number, palette_group_index)

    return bg_color_for_palette_group(palette_group)


def bg_color_for_palette_group(palette_group: PaletteGroup) -> QColor:
    return NESPalette[palette_group[0][0]]



from math import ceil

from PySide6.QtCore import QPoint, QRect, QSize
from PySide6.QtGui import (
    QBrush,
    QCloseEvent,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QResizeEvent,
)
from PySide6.QtWidgets import QComboBox, QLabel, QLayout, QStatusBar, QToolBar, QWidget

from foundry import icon
from foundry.game.File import ROM
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.GraphicsSet import GraphicsSet
from foundry.game.gfx.Palette import (
    PALETTE_GROUPS_PER_OBJECT_SET,
    bg_color_for_object_set,
    load_palette_group,
)
from foundry.gui.CustomChildWindow import CustomChildWindow
from foundry.gui.LevelSelector import OBJECT_SET_ITEMS
from foundry.gui.Spinner import Spinner


class BlockViewer(CustomChildWindow):
    def __init__(self, parent):
        super(BlockViewer, self).__init__(parent, "Block Viewer")

        self._object_set = 0
        self.sprite_bank = BlockBank(parent=self)

        self.setCentralWidget(self.sprite_bank)

        self.toolbar = QToolBar(self)

        self.prev_os_action = self.toolbar.addAction(icon("arrow-left.svg"), "Previous object set")
        self.prev_os_action.triggered.connect(self.prev_object_set)

        self.next_os_action = self.toolbar.addAction(icon("arrow-right.svg"), "Next object set")
        self.next_os_action.triggered.connect(self.next_object_set)

        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_out_action.triggered.connect(self.sprite_bank.zoom_out)

        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")
        self.zoom_in_action.triggered.connect(self.sprite_bank.zoom_in)

        self.bank_dropdown = QComboBox(parent=self.toolbar)
        self.bank_dropdown.addItems(OBJECT_SET_ITEMS)
        self.bank_dropdown.setCurrentIndex(0)

        self.bank_dropdown.currentIndexChanged.connect(self.on_combo)

        self.palette_group_spinner = Spinner(self, maximum=PALETTE_GROUPS_PER_OBJECT_SET - 1, base=10)
        self.palette_group_spinner.valueChanged.connect(self.on_palette)

        self.toolbar.addWidget(self.bank_dropdown)
        self.toolbar.addWidget(QLabel(" Object Palette: "))
        self.toolbar.addWidget(self.palette_group_spinner)

        self.addToolBar(self.toolbar)

        self.layout().setSizeConstraint(QLayout.SetFixedSize)

        self.setStatusBar(QStatusBar(self))

        return

    def closeEvent(self, event: QCloseEvent):
        self.toolbar.close()
        super().closeEvent(event)

    @property
    def object_set(self):
        return self._object_set

    @object_set.setter
    def object_set(self, value):
        self._object_set = value

        self._after_object_set()

    @property
    def palette_group(self):
        return self.palette_group_spinner.value()

    @palette_group.setter
    def palette_group(self, value):
        self.palette_group_spinner.setValue(value)

    def prev_object_set(self):
        self.object_set = max(self.object_set - 1, 0)

    def next_object_set(self):
        self.object_set = min(self.object_set + 1, 0xE)

    def _after_object_set(self):
        self.sprite_bank.object_set = self.object_set

        self.bank_dropdown.setCurrentIndex(self.object_set)

        self.sprite_bank.update()

    def on_combo(self, _):
        self.object_set = self.bank_dropdown.currentIndex()

        self.sprite_bank.object_set = self.object_set

        self.sprite_bank.update()

    def on_palette(self, value):
        self.sprite_bank.palette_group = value
        self.sprite_bank.update()


class BlockBank(QWidget):
    def __init__(self, parent, object_set=0, palette_group=0, zoom=2):
        super(BlockBank, self).__init__(parent)
        self.setMouseTracking(True)

        self.sprites = 256
        self.zoom_step = 256
        self.sprites_horiz = 16
        self.sprites_vert = ceil(self.sprites / self.sprites_horiz)

        self.object_set = object_set
        self.palette_group = palette_group
        self.zoom = zoom

        self._size = QSize(self.sprites_horiz * Block.WIDTH * self.zoom, self.sprites_vert * Block.HEIGHT * self.zoom)

        self.setFixedSize(self._size)

    def resizeEvent(self, event: QResizeEvent):
        self.update()

    def zoom_in(self):
        self.zoom += 1
        self._after_zoom()

    def zoom_out(self):
        self.zoom = max(self.zoom - 1, 1)
        self._after_zoom()

    def _after_zoom(self):
        new_size = QSize(self.sprites_horiz * Block.WIDTH * self.zoom, self.sprites_vert * Block.HEIGHT * self.zoom)

        self.setFixedSize(new_size)

    def mouseMoveEvent(self, event: QMouseEvent):
        x, y = event.pos().toTuple()

        block_length = Block.WIDTH * self.zoom

        column = x // block_length
        row = y // block_length

        dec_index = row * self.sprites_horiz + column
        hex_index = hex(dec_index).upper().replace("X", "x")

        status_message = f"Row: {row}, Column: {column}, Index: {dec_index} / {hex_index}"

        self.parent().statusBar().showMessage(status_message)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        bg_color = bg_color_for_object_set(self.object_set, 0)
        painter.setBrush(QBrush(bg_color))

        painter.drawRect(QRect(QPoint(0, 0), self.size()))

        graphics_set = GraphicsSet(self.object_set)
        palette = load_palette_group(self.object_set, self.palette_group)
        tsa_data = ROM.get_tsa_data(self.object_set)

        horizontal = self.sprites_horiz

        block_length = Block.WIDTH * self.zoom

        for i in range(self.sprites):
            block = Block(i, palette, graphics_set, tsa_data)

            x = (i % horizontal) * block_length
            y = (i // horizontal) * block_length

            block.draw(painter, x, y, block_length)

        return
